unit hiBigBitmapViewer;

interface

uses
  Kol,Windows,Share,Debug,Win,Messages;

type
  THiBigBitmapViewer = class(THIWin)
    private
      FBmp: KOL.PBitmap;
      FFileName: string;
      FWidthBmp: integer;
      FHeightBmp: integer;
      FBufSize: integer;
      OldMes: TOnMessage;
      FCheckSize: boolean;
      FCancel: boolean;
      function GetDIBInBands( const FileName: string;
        DestBitmap: KOL.PBitmap; BufferSize: integer;
        out TotalBitmapWidth, TotalBitmapHeight: Integer ): Boolean;
      procedure Paint(Sender: PControl; DC: HDC);
      procedure Resize(Obj: PObj);
      function onMessage(var Msg: TMsg; var Rslt: Integer): Boolean;
    public
      _prop_FileName: string;
      _prop_BufSize: integer;
      _data_FileName: THI_Event;
      _event_onEndLoad: THI_Event;
      _event_onLoadProgress: THI_Event;

      constructor Create(Parent: PControl);
      procedure Init;
      procedure Load;
      procedure _work_doLoadBigBitmap(var _Data: TData; Index: word);
      procedure _work_doClear(var _Data: TData; Index: word);
      procedure _work_doBufSize(var _Data: TData; Index: word);
      procedure _work_doCancel(var _Data: TData; Index: word);
      procedure _var_Bitmap(var _Data: TData; Index: word);
      procedure _var_BmpWidth(var _Data: TData; Index: word);
      procedure _var_BmpHeight(var _Data: TData; Index: word);
   end;

implementation

{$R-}

function MyGetMem( size: DWORD ): pointer;
begin
  result := Pointer(GlobalAlloc(GPTR, size));
end;

procedure MyFreeMem( p: pointer );
begin
  if p = Nil then Exit;
  GlobalFree(THandle(p));
end;

procedure THiBigBitmapViewer._work_doCancel;
begin
  FCancel := True;
end;

constructor THiBigBitmapViewer.Create;
begin
  inherited;
  InitAdd(Load);
end;

procedure THiBigBitmapViewer.Load;
var
  dt: TData;
begin
  FFileName := ReadString(dt,_data_FileName,_prop_FileName);
  FBufSize := _prop_BufSize;
  FCheckSize := True;
  if FileExists(FFileName) then
    if not GetDIBInBands(FFileName, FBmp, FBufSize, FWidthBmp, FHeightBmp) then
      _Debug('Error: load bitmap ' + FFileName);
end;

function THiBigBitmapViewer.onMessage(var Msg: TMsg; var Rslt: Integer): Boolean;
begin
  case Msg.message of
    WM_EXITSIZEMOVE:
    begin
      if (FileExists(FFileName)) and (FCheckSize = True) then
      begin
        if not GetDIBInBands(FFileName, FBmp, FBufSize, FWidthBmp, FHeightBmp) then
          _Debug('Error: load bitmap ' + FFileName);
        FCheckSize := False;
        Control.Invalidate;
      end;
    end;

    WM_MOVE: FCheckSize := False;

    WM_SYSCOMMAND:
    begin
      if (Msg.wParam = SC_MAXIMIZE) or (Msg.wParam = SC_RESTORE) then
      begin
        FCheckSize := True;
        PostMessage(Control.ParentForm.Handle,WM_EXITSIZEMOVE,0,0);
      end;
    end;
  end;

  Result := _hi_OnMessage(OldMes,Msg,Rslt);
end;

procedure THiBigBitmapViewer.Resize(Obj: PObj);
begin
  FCheckSize := True;
  FBmp.Width := Control.Width;
  FBmp.Height := Control.Height;
  if not FileExists(FFileName) then
  begin
    FBmp.Canvas.Brush.Color := _prop_Color;
    FBmp.Canvas.Brush.BrushStyle := bsSolid;
    FBmp.Canvas.FillRect(MakeRect(0,0, FBmp.Width, FBmp.Height));
  end;
  Control.Invalidate;
end;

procedure THiBigBitmapViewer.Init;
begin
  Control := NewPaintBox(FParent);
inherited;
  FBmp := NewBitmap(Control.Width, Control.Height);
  FBmp.Width := Control.Width;
  FBmp.Height := Control.Height;
  FBmp.PixelFormat:= pf24Bit;
  FBmp.Canvas.Brush.Color := _prop_Color;
  FBmp.Canvas.Brush.BrushStyle := bsSolid;
  FBmp.Canvas.FillRect(MakeRect(0,0, FBmp.Width, FBmp.Height));

  OldMes := Control.ParentForm.OnMessage;
  Control.ParentForm.OnMessage := OnMessage;

  Control.Show;
  Control.OnResize := Resize;
  Control.OnPaint := Paint;
  Control.Invalidate;
  PostMessage(Control.ParentForm.Handle,WM_EXITSIZEMOVE,0,0);
end;

procedure THiBigBitmapViewer._work_doLoadBigBitmap;
begin
  FFileName := ReadString(_Data, _data_FileName);
  if FileExists(FFileName) then
    if not GetDIBInBands(FFileName, FBmp, FBufSize, FWidthBmp, FHeightBmp) then
      _Debug('Error: load bitmap ' + FFileName);
  Control.Invalidate;
end;

procedure THiBigBitmapViewer._work_doClear;
begin
  FBmp.Canvas.Brush.Color := _prop_Color;
  FBmp.Canvas.Brush.BrushStyle := bsSolid;
  FBmp.Canvas.FillRect(MakeRect(0,0, FBmp.Width, FBmp.Height));
  FFileName := '';
  FWidthBmp := 0;
  FHeightBmp := 0;
  Control.Invalidate;
  _hi_onEvent(_event_onEndLoad);
end;

procedure THiBigBitmapViewer._work_doBufSize;
begin
  FBufSize := ToInteger(_Data);
end;

procedure THiBigBitmapViewer._var_Bitmap;
begin
  dtBitmap(_Data,FBmp);
end;

procedure THiBigBitmapViewer._var_BmpWidth;
begin
  dtInteger(_Data,FWidthBmp);
end;

procedure THiBigBitmapViewer._var_BmpHeight;
begin
  dtInteger(_Data,FHeightBmp);
end;

procedure THiBigBitmapViewer.Paint(Sender: PControl; DC: HDC);
begin
  FBmp.Draw(Control.Canvas.Handle, 0,0);
end;

{ This code will fill a bitmap by resampling an image coming from a big bitmap on disk.

  FileName.- Name of the uncompressed bitmap to read
  DestBitmap.- Target bitmap  where the bitmap on disk will be resampled.
  BufferSize.- The size of a memory buffer used for reading scanlines from the physical bitmap on disk.
    This value will decide how many scanlines can be read from disk at the same time, with always a
    minimum value of 2 scanlines.

  Will return false on error.
}
function THiBigBitmapViewer.GetDIBInBands( const FileName: string;
  DestBitmap: KOL.PBitmap; BufferSize: integer;
  out TotalBitmapWidth, TotalBitmapHeight: integer): boolean;
var
  FileSize: integer;            // calculated file size
  ImageSize: integer;           // calculated image size
  dest_MaxScans: integer;       // number of scanline from source bitmap
  dsty_top: integer;            // used to calculate number of passes
  NumPasses: integer;           // number of passed needed
  dest_Residual: integer;       // number of scanlines on last band
  Stream: PStream;              // stream used for opening the bitmap
  bmf: TBITMAPFILEHEADER;       // the bitmap header
  lpBitmapInfo: PBITMAPINFO;    // bitmap info record
  BitmapHeaderSize: integer;    // size of header of bitmap
  SourceIsTopDown: boolean;     // is reversed bitmap ?
  SourceBytesPerScanLine: integer;  // number of bytes per scanline
  SourceLastScanLine: Extended;     // last scanline processes
  SourceBandHeight: Extended;       //
  BitmapInfo: PBITMAPINFO;
  img_start: integer;
  img_end: integer;
  img_numscans: integer;
  OffsetInFile: integer;
  OldHeight: integer;
  bits: Pointer;
  CurrentTop: integer;
  CurrentBottom: integer;
  Pos,Pos2: integer;
begin
  Result := False;
  // open the big bitmap
  Stream := NewReadFileStream(FileName);
  // total size of bitmap
  FileSize := Stream.Size;
  // read the header
  Stream.Read( bmf, sizeof( TBITMAPFILEHEADER ) );
  // calculate header size
  BitmapHeaderSize := bmf.bfOffBits - sizeof( TBITMAPFILEHEADER );
  // calculate size of bitmap bits
  ImageSize := FileSize - integer( bmf.bfOffBits );
  // check for valid bitmap and exit if not
  if ( ( bmf.bfType <> $4D42 ) Or
    ( integer( bmf.bfOffBits ) < 1 ) Or
    ( FileSize < 1 ) Or ( BitmapHeaderSize < 1 ) Or ( ImageSize < 1 ) Or
    ( FileSize < ( SizeOf( TBITMAPFILEHEADER ) + BitmapHeaderSize + ImageSize ) ) ) then
  begin
    Stream.Free;
    Exit;
  end;
  lpBitmapInfo := MyGetMem( BitmapHeaderSize );
  try
    Stream.Read( lpBitmapInfo^, BitmapHeaderSize );
    // check for uncompressed bitmap
    if ( ( lpBitmapInfo^.bmiHeader.biCompression = BI_RLE4 ) Or
      ( lpBitmapInfo^.bmiHeader.biCompression = BI_RLE8 ) ) then
    begin
      Exit;
    end;

    // bitmap dimensions
    TotalBitmapWidth := lpBitmapInfo^.bmiHeader.biWidth;
    TotalBitmapHeight := abs( lpBitmapInfo^.bmiHeader.biHeight );

    // is reversed order ?
    SourceIsTopDown := ( lpBitmapInfo^.bmiHeader.biHeight < 0 );

    // calculate number of bytes used per scanline
    SourceBytesPerScanLine := ((((lpBitmapInfo^.bmiHeader.biWidth *
      lpBitmapInfo^.bmiHeader.biBitCount) + 31) And Not 31) Div 8);

    // adjust buffer size
    if BufferSize < Abs( SourceBytesPerScanLine ) then
      BufferSize := Abs( SourceBytesPerScanLine );

    // calculate number of scanlines for every pass on the destination bitmap
    dest_MaxScans := round( BufferSize / abs( SourceBytesPerScanLine ) );
    dest_MaxScans := round( dest_MaxScans * ( DestBitmap.Height / TotalBitmapHeight ) );

    if dest_MaxScans < 2 then
      dest_MaxScans := 2;         // at least two scan lines

    // is not big enough ?
    if dest_MaxScans > TotalBitmapHeight then
      dest_MaxScans := TotalBitmapHeight;

    { count the number of passes needed to fill the destination bitmap }
    dsty_top := 0;
    NumPasses := 0;
    while ( dsty_Top + dest_MaxScans ) <= DestBitmap.Height Do
    begin
      inc( NumPasses );
      inc( dsty_top, dest_MaxScans );
    end;
    if NumPasses = 0 then Exit;

    // calculate scanlines on last pass
    dest_Residual := DestBitmap.Height Mod dest_MaxScans;

    // now calculate how many scanlines in source bitmap needed for every band on the destination bitmap
    SourceBandHeight :=( TotalBitmapHeight * ( 1 - ( dest_Residual / DestBitmap.Height ) ) ) / NumPasses;

    // initialize first band
    CurrentTop := 0;
    CurrentBottom := dest_MaxScans;

    // a floating point used in order to not loose last scanline precision on source bitmap
    // because every band on target could be a fraction (not integral) on the source bitmap
    SourceLastScanLine := 0.0;

    Pos := 0; Pos2 := 0;

    while CurrentTop < DestBitmap.Height Do
    begin
      Inc(Pos,dest_MaxScans);
      Pos2 := Trunc((Pos/DestBitmap.Height)*100);
      if (Pos2 mod 5) = 0 then
      begin
        _hi_OnEvent(_event_onLoadProgress,Pos2);
        Applet.ProcessMessages;
      end;

      if FCancel = True then
      begin
        FCancel := False;
        Stream.Free;
        MyFreeMem( lpBitmapInfo );
        _hi_onEvent(_event_onEndLoad);
        _hi_OnEvent(_event_onLoadProgress,0);
        FBmp.Canvas.Brush.Color := _prop_Color;
        FBmp.Canvas.Brush.BrushStyle := bsSolid;
        FBmp.Canvas.FillRect(MakeRect(0,0, FBmp.Width, FBmp.Height));
        FFileName := '';
        Result:= True;
        Control.Invalidate;
        Exit;
      end;

      // scanline start of band in source bitmap
      img_start := Round( SourceLastScanLine );
      SourceLastScanLine := SourceLastScanLine + SourceBandHeight;
      // scanline finish of band in source bitmap
      img_end := Round( SourceLastScanLine );
      if img_end > TotalBitmapHeight - 1 then
        img_end := TotalBitmapHeight - 1;
      img_numscans := img_end - img_start;
      if img_numscans < 1 then Break;
      OldHeight := lpBitmapInfo^.bmiHeader.biHeight;
      if SourceIsTopDown then
        lpBitmapInfo^.bmiHeader.biHeight := -img_numscans
      else
        lpBitmapInfo^.bmiHeader.biHeight := img_numscans;

      // memory used to read only the current band
      bits := MyGetMem( Abs(SourceBytesPerScanLine) * img_numscans);

      try
        // calculate offset of band on disk
        OffsetInFile := TotalBitmapHeight - (img_start + img_numscans);
        Stream.Seek( integer( bmf.bfOffBits ) + ( OffsetInFile * abs( SourceBytesPerScanLine ) ), spBegin );
        Stream.Read( bits^, abs( SourceBytesPerScanLine ) * img_numscans );

        SetStretchBltMode( DestBitmap.Canvas.Handle, COLORONCOLOR );
        // now stretch the band readed to the destination bitmap
        StretchDIBits(
          DestBitmap.Canvas.Handle,
          0,
          CurrentTop,
          DestBitmap.Width,
          Abs(CurrentBottom - CurrentTop),
          0,
          0,
          TotalBitmapWidth,
          img_numscans,
          Bits,
          lpBitmapInfo^,
          DIB_RGB_COLORS, SRCCOPY );
      finally
        MyFreeMem( bits );
        lpBitmapInfo^.bmiHeader.biHeight := OldHeight;
      end;

      CurrentTop := CurrentBottom;
      CurrentBottom := CurrentTop + dest_MaxScans;
      if CurrentBottom > DestBitmap.Height then
        CurrentBottom := DestBitmap.Height;

    end;
  finally
    Stream.Free;
    MyFreeMem( lpBitmapInfo );
  end;
  Result:= True;
  _hi_onEvent(_event_onEndLoad);
  _hi_OnEvent(_event_onLoadProgress,0);
end;


end.
